﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Net.Sockets;
using System.Threading;
using System.Net;
using System.Collections;
using System.IO;
using System.Drawing;
using System.ComponentModel;

namespace ThunderDashboard
{
    // Used Asynchronous Client Socket Example from MSDN and WebcamViewerExtensionCompat.java from GitHub


    class CameraControl : PictureBox
    {
        // delegate that gets called when image has been processed
        // targetInfo - each target's info
        // targets    - how many targets are in targetInfo
        // imageWidth - the width of the entire image
        // imageHeight - the height of the entire image
        //
        // caller returns true if they want the targets to be drawn
        public delegate bool VisionDelegate(ThunderVision.TARGET_INFO[] targetInfo, int targets, int imageWidth, int imageHeight);

        // camera resolution
        public enum Resolution
        {
            ResolutionDoNotSet = 99,
            Resolution640x480 = 0,
            Resolution320x240 = 1,
            Resolution160x120 = 2,
        };

        private string FPS(int fps)
        {
            return (string.Format("{0} fps", fps));
        }

        // constructor
        public CameraControl()
        {
            _started = false;

            // create control for resolutions
            _resChoices = new ComboBox();
            _resChoices.DropDownStyle = ComboBoxStyle.DropDownList;
            _resChoices.SelectedIndexChanged += _resChoices_SelectedIndexChanged;
            _resChoices.Size = new Size(76, 21);
            _resChoices.Location = new Point(0, Bottom - _resChoices.Height);
            _resChoices.Anchor = AnchorStyles.Left | AnchorStyles.Bottom;
            _resChoices.Items.Add("640 x 480");
            _resChoices.Items.Add("320 x 240");
            _resChoices.Items.Add("160 x 120");
            AddToolTip(_resChoices, "Resolution");
            Controls.Add(_resChoices);

            // create control for frames per second
            _fpsChoices = CreateToolbarDropdown(null, _fpsChoices_SelectedIndexChanged, "Frame Rate", new string[] { FPS(30), FPS(25), FPS(20), FPS(15), FPS(10), FPS(5) }, FPS(Properties.Settings.Default.CameraFrameRate), 56);
            _fpsChoices.Location = new Point(Right - _fpsChoices.Width, Bottom - _fpsChoices.Height);
            _fpsChoices.Anchor = AnchorStyles.Right | AnchorStyles.Bottom;
            Controls.Add(_fpsChoices);

            // add brightness and exposure panel
            _panelBright = new TableLayoutPanel();
            _panelBright.RowStyles.Add(new RowStyle(SizeType.AutoSize));
            _panelBright.Anchor = AnchorStyles.Right | AnchorStyles.Top;
            _panelBright.AutoSize = true;
            _panelBright.AutoSizeMode = AutoSizeMode.GrowAndShrink;
            _panelBright.BackColor = Color.FromArgb(207, 214, 229);
            _panelBright.Padding = new Padding(0);
            _panelBright.Margin = new Padding(0);
            _defaults = CreateToolbarButton(_panelBright, Properties.Resources.Defaults, _defaults_Click, "Set 2017 defaults");
            //CreateToolbarText(_panelBright, "B:");
            _brightness = CreateToolbarNumericUpDown(_panelBright, _brightness_ValueChanged, "Brightness", 0, 100, Properties.Settings.Default.CameraBrightness);
            //CreateToolbarText(_panelBright, "E:");
            _exposure = CreateToolbarNumericUpDown(_panelBright, _exposure_ValueChanged, "Exposure", -1, 100, Properties.Settings.Default.CameraExposure);
            _whiteBalance = CreateToolbarDropdown(_panelBright, _whiteBalance_SelectedIndexChanged, "White Balance", new string[] { "auto", "fi", "fo1", "fo2", "ff1", "ff2" }, Properties.Settings.Default.CameraWhiteBalance, 44);
            _panelBright.Location = new Point(Right - _panelBright.Width, 0);
            Controls.Add(_panelBright);

            // add image toolbar panel
            _panelImage = new TableLayoutPanel();
            _panelImage.AutoSize = true;
            _panelImage.AutoSizeMode = AutoSizeMode.GrowAndShrink;
            _panelImage.BackColor = Color.FromArgb(207, 214, 229);
            _panelImage.RowStyles.Add(new RowStyle(SizeType.AutoSize));
            _panelImage.Anchor = AnchorStyles.Left | AnchorStyles.Top;
            _panelImage.Padding = new Padding(0);
            _playButton = CreateToolbarButton(_panelImage, Properties.Resources.Pause, _playButton_Click, "Pause live stream", true);
            _enableButton = CreateToolbarButton(_panelImage, Properties.Resources.CameraOff, _enableButton_Click, "Stop camera connection", true);
            _liveProcessButton = CreateToolbarButton(_panelImage, Properties.Resources.LiveProcessOn, _liveProcessButton_Click, "Start live processing, even if robot isn't asking", true);
            _saveButton = CreateToolbarButton(_panelImage, Properties.Resources.Save, _saveButton_Click, "Save current image to a bitmap file");
            _loadButton = CreateToolbarButton(_panelImage, Properties.Resources.Open, _loadButton_Click, "Load bitmap file to be current image");
            _processButton = CreateToolbarButton(_panelImage, Properties.Resources.Process, _processButton_Click, "Process currently displayed image", true);
            _panelImage.Location = new Point(0, 0);
            Controls.Add(_panelImage);

            SetPlayMode(true);
        }


        private Button CreateToolbarButton(TableLayoutPanel panel, Image image, EventHandler handler, string toolTip, bool separator = false)
        {
            Button control = new Button();
            control.AutoSize = true;
            control.AutoSizeMode = AutoSizeMode.GrowAndShrink;
            control.ImageAlign = ContentAlignment.MiddleCenter;
            control.FlatStyle = FlatStyle.Flat;
            control.FlatAppearance.BorderSize = 0;
            control.Margin = new Padding(0, 0, (separator ? 10 : 2), 0);
            control.Image = image;
            control.Click += handler;


            if (panel != null)
            {
                panel.ColumnStyles.Add(new ColumnStyle(SizeType.AutoSize));
                panel.Controls.Add(control, panel.Controls.Count, 0);
            }

            AddToolTip(control, toolTip);
            return (control);
        }
        private NumericUpDown CreateToolbarNumericUpDown(TableLayoutPanel panel, EventHandler handler, string toolTip, decimal min = 0, decimal max = 100, decimal value = 100, int width = 40)
        {
            NumericUpDown control = new NumericUpDown();
            control.Minimum = min;
            control.Maximum = max;
            control.Margin = new Padding(0, 0, 2, 0);
            control.Value = value;
            control.Width = width;
            control.ValueChanged += handler;


            if (panel != null)
            {
                panel.ColumnStyles.Add(new ColumnStyle(SizeType.AutoSize));
                panel.Controls.Add(control, panel.Controls.Count, 0);
            }

            AddToolTip(control, toolTip);
            return (control);
        }
        private Label CreateToolbarText(TableLayoutPanel panel, string text)
        {
            Label control = new Label();
            control.AutoSize = true;
            control.Margin = new Padding(0, 2, 0, 0);
            control.Text = text;


            if (panel != null)
            {
                panel.ColumnStyles.Add(new ColumnStyle(SizeType.AutoSize));
                panel.Controls.Add(control, panel.Controls.Count, 0);
            }

            //AddToolTip(control, toolTip);
            return (control);
        }
        private ComboBox CreateToolbarDropdown(TableLayoutPanel panel, EventHandler handler, string toolTip, string[] list, string value, int width = 56)
        {
            ComboBox control = new ComboBox();
            control.DropDownStyle = ComboBoxStyle.DropDownList;
            control.Margin = new Padding(0, 0, 2, 0);
            control.SelectedIndexChanged += handler;
            control.Size = new Size(width, 21);
            foreach (string item in list)
            {
                control.Items.Add(item);
                if (item == value)
                {
                    control.SelectedItem = value;
                }
            }

            if (panel != null)
            {
                panel.ColumnStyles.Add(new ColumnStyle(SizeType.AutoSize));
                panel.Controls.Add(control, panel.Controls.Count, 0);
            }

            AddToolTip(control, toolTip);
            return (control);
        }
        private void AddToolTip(Control control, string toolTip)
        {
            // set tool tip
            if (!string.IsNullOrEmpty(toolTip))
            {
                if (_tooltip == null)
                {
                    _tooltip = new ToolTip();
                    _tooltip.ShowAlways = true;
                }
                _tooltip.SetToolTip(control, toolTip);
            }
        }
        public bool Testing
        {
            set
            {
                //_loadButton.Visible = value;
                //_processButton.Visible = value;
            }
        }

        // load image
        private void _loadButton_Click(object sender, EventArgs e)
        {
            OpenFileDialog ofd = new OpenFileDialog();
            ofd.CheckFileExists = true;
            ofd.CheckPathExists = true;
            ofd.DefaultExt = ".bmp";
            ofd.Filter = "Bitmap Files(*.BMP)| *.BMP";
            //ofd.InitialDirectory = "C:\\Users\\Robotics\\Desktop";
            if (DialogResult.OK == ofd.ShowDialog())
            {
                try
                {
                    _loadedFile = ofd.FileName;
                    this.Image = new Bitmap(ofd.FileName);
                }
                catch
                {
                    MessageBox.Show("Unable to open the image");
                }
            }
        }
        // stop/start
        private void _enableButton_Click(object sender, EventArgs e)
        {
            _allowCamera = !_allowCamera;
            _enableButton.Image = (_allowCamera ? Properties.Resources.CameraOff : Properties.Resources.CameraOn);
            _tooltip.SetToolTip(_enableButton, (_allowCamera ? "Stop camera connection" : "Start camera connection"));
            if (_allowCamera)
            {
                this.Image = null;  // remove 'disabled' image
                Start();
            }
            else
            {
                this.Image = Properties.Resources.CameraDisabled;
                Stop();
            }
        }
        // play/pause
        private void _playButton_Click(object sender, EventArgs e)
        {
            SetPlayMode(!_updateDisplay);
        }
        private void SetPlayMode(bool play)
        {
            _updateDisplay = play;
            _playButton.Image = (_updateDisplay ? Properties.Resources.Pause : Properties.Resources.Play);
            _tooltip.SetToolTip(_playButton, (_updateDisplay ? "Pause live stream" : "Restart live stream"));
            _saveButton.Enabled = !_updateDisplay;
            _loadButton.Enabled = !_updateDisplay;
            _processButton.Enabled = !_updateDisplay;
            _liveProcessButton.Enabled = _updateDisplay;
        }
        // save
        private void _saveButton_Click(object sender, EventArgs e)
        {
            SaveFileDialog sfd = new SaveFileDialog();
            sfd.AddExtension = true;
            sfd.CheckPathExists = true;
            sfd.OverwritePrompt = true;
            sfd.CheckPathExists = true;
            sfd.DefaultExt = ".bmp";
            sfd.Filter = "Bitmap Files(*.BMP)| *.BMP";
            //sfd.InitialDirectory = "C:\\Users\\Robotics\\Desktop";
            if (DialogResult.OK == sfd.ShowDialog())
            {
                try
                {
                    this.Image.Save(sfd.FileName, System.Drawing.Imaging.ImageFormat.Bmp);
                }
                catch
                {
                    MessageBox.Show("Unable to save the image");
                }
            }
        }
        // process image
        private void _processButton_Click(object sender, EventArgs e)
        {
            //if (!string.IsNullOrEmpty(_file))
            {
                ProcessTarget(_loadedFile);
            }
        }
        // live processing of images
        private void _liveProcessButton_Click(object sender, EventArgs e)
        {
            _liveProcessing = !_liveProcessing;
            _liveProcessButton.Image = (_liveProcessing ? Properties.Resources.LiveProcessOff : Properties.Resources.LiveProcessOn);
            _tooltip.SetToolTip(_liveProcessButton, (_liveProcessing ? "Stop live processing, only do it when robot asks" : "Start live processing, even if robot isn't asking"));
        }

        // start and stop vision processing
        public void VisionStart()
        {
            _visionProcessing = (_visionDelegate != null);
        }
        public void VisionStop()
        {
            _visionProcessing = false;
        }

        // start camera processing
        public void Start()
        {
            Start(_cameraIp, _port, _statusTextBox, _cameraFlip, _visionDelegate);
        }
        public void Start(string ipAddress, int port /*= 1180*/, TextBox status /*= null*/, bool flip /*= false*/, VisionDelegate visionDelegate)
        {
            if (!_allowCamera)
            {
                return;
            }
            if (_started)
            {
                Stop();
                Thread.Sleep(250); // wait for current camera to stop
                _started = false;
            }
            _port = port;
            _cameraIp = ipAddress;
            _cameraFlip = flip;
            _statusTextBox = status;
            _visionDelegate = visionDelegate;

            // get resolution and frame rate
            try
            {
                _resChoices.SelectedIndex = Properties.Settings.Default.CameraResolution;
                _fpsChoices.SelectedItem = FPS(Properties.Settings.Default.CameraFrameRate);
                _brightness.Value = Properties.Settings.Default.CameraBrightness;
                _exposure.Value = Properties.Settings.Default.CameraExposure;
            }
            catch
            {
                _resChoices.SelectedIndex = 0;
                _fpsChoices.SelectedIndex = 0;
                _brightness.Value = 50;
                _exposure.Value = 50;
            }
            UpdateRobot();  // make sure the robot is in sync

            // only create background worker if not already running
            if (_bw == null)
            {
                _bw = new BackgroundWorker();
                _bw.DoWork += new DoWorkEventHandler(StartConnection);
                _bw.RunWorkerAsync();
            }

            _started = true;
        }

        // reset the cammera (close thread and restart it)
        public void Restart()
        {
            if ((_bw != null) && _started)
            {
                UpdateStatusDescription("restarting");
                _resetCamera = true;
            }
        }

        // called when the network tables is disconnected
        public void Disconnected()
        {
            if (client != null)
            {
                client.Close();
            }
        }

        // called when network tables are re-connected
        public void Connected()
        {
            UpdateRobot();  // make sure the robot is in sync
        }

        // stop using camera
        public void Stop()
		{
            if (!_started)
            {
                if (!_allowCamera)
                {
                    this.Image = Properties.Resources.CameraDisabled;
                }
                return;
            }
            //if (client != null)
            //{
            //    client.Close();
            //    statusTextBox.Text = "Socket Shutdown";
            //}
            _resetCamera = false; // do not restart it
            _shutdownCamera = true;
		}

        // set the resolution for the camera
        private void SetResolution(Resolution resolution)
        {
            // if the rate didn't change, then just return
            if (_resolution == resolution)
            {
                return;
            }

            // save the new value and refresh the camera
            _resolution = resolution;
            Restart();
        }

        // set the frames per second for the camera
        private void SetFrameRate(int fps)
        {
            // if the rate didn't change, then just return
            if (_fps == fps)
            {
                return;
            }

            // save the new value and refresh the camera
            _fps = fps;
            Restart();
        }

        // set all network table based values
        private void UpdateRobot()
        {
            DashboardNetworkTable.PutNumber("camera_brightness", (double)_brightness.Value);
            DashboardNetworkTable.PutNumber("camera_exposure", (double)_exposure.Value);
            if (_whiteBalance.SelectedIndex >= 0)
            {
                DashboardNetworkTable.PutString("camera_whitebalance", (_whiteBalance.SelectedItem as string));
            }
        }

        // State object for receiving data from remote camera.
        private class StateObject
		{
			public Socket workSocket = null;    // clilent socket
			public byte[] buffer;               // receive buffer
			public int bytesRequested = 0;
			public int bytesReceived = 0;
		}

        // background worker thread
        private void StartConnection(object obj, DoWorkEventArgs eEventArgs)
        {
            huffman_table = new byte[huffman_table_int.Length];
            for (int i = 0; i < huffman_table.Length; ++i)
            {
                huffman_table[i] = (byte)huffman_table_int[i];
            }

            // retry if we are reseting/changing settings
            do
            {
                _resetCamera = false;   // clear flag so we only reset once

                // see if the robot is even there
                if (Program.PingRobot(_cameraIp))
                {
                    RunCamera();
                    UpdateStatusDescription("waiting to retry camera connection");
                }

                // robot is not there, so wait 3 seconds and try again. but wait
                //in 100ms intervals in case user closes the form while we were waiting
                for (int i = 0; ((i < 30) && !_shutdownCamera && !_resetCamera); i++)
                {
                    Thread.Sleep(100);
                }

            } while (!_shutdownCamera);

            UpdateStatusDescription("camera ended..."); // init status window
            if (!_allowCamera)
            {
                // camera feed was disabled, so show image indicating such.
                // we do this here because an image might have come in when we shut it down
                if (InvokeRequired)
                {
                    BeginInvoke(new UpdateImageDelegate(UpdateImage), new byte[] { 0 });
                }
                else
                {
                    UpdateImage(null);
                }
            }
        }

        // runs the camera (for background worker) until told to exit camera
        // Returns: true if we should wait and retry again
        private bool RunCamera()
        {
            bool ret = false;

            client = null;					// make sure client is null
            _shutdownCamera = false;		// init it once
            UpdateStatusDescription("connecting...");	// init status window
            while (!_shutdownCamera && !_resetCamera)		// needed incase we are stuck in outer try/catch
            {
                // Connect to remote camera.
                try
                {
                    // Establish the remote endpoint for the socket.
                    //Thread.Sleep(5000);
                    IPHostEntry ipHostInfo = Dns.GetHostEntry(_cameraIp);

                    foreach (IPAddress ipAddress in ipHostInfo.AddressList)
                    {
                        // We get back two addresses: IPv6 and IPv4. Only the IPv4 works for the
                        // socket connection
                        if (ipAddress.AddressFamily == AddressFamily.InterNetwork)
                        {
                            IPEndPoint remoteEP = new IPEndPoint(ipAddress, _port);

                            // Create a TCP/IP socket.
                            client = new Socket(AddressFamily.InterNetwork,
                                SocketType.Stream, ProtocolType.Tcp);
                            if (client == null)
                            {
                                //UpdateStatusDescription("socket fail");
                                break;
                            }

                            // Connect to the remote endpoint.
                            //UpdateStatusDescription("try connect");
                            connectDone.Reset();
                            client.BeginConnect(remoteEP, new AsyncCallback(ConnectCallback), client);
                            connectDone.WaitOne();
                            if (connectError)
                            {
                                //UpdateStatusDescription("connect fail");
                                client.Close();
                                client = null;

                                // break out and start all over looking for the robot
                                _resetCamera = true;
                                break;
                            }

                            // configure the camera
                            if (_resolution != Resolution.ResolutionDoNotSet)
                            {
                                int res = 0;
                                switch (_resolution) // frame size: 0=640x480, 1=320x240, 2=160x120
                                {
                                    case Resolution.Resolution640x480: res = 0; break;
                                    case Resolution.Resolution320x240: res = 1; break;
                                    case Resolution.Resolution160x120: res = 2; break;
                                }
                                // Send data to the roboRIO
                                // (frames per second, hardware compression, frame size)
                                sendDone.Reset();
                                if (!Send(client, new int[] { _fps, -1, res }))  
                                {
                                    //UpdateStatusDescription("send fail");
                                    //client.Shutdown(SocketShutdown.Both);
                                    //client.Close();
                                    client = null;
                                    break;
                                }
                                sendDone.WaitOne();
                                if (sendError)
                                {
                                    //UpdateStatusDescription("send done fail");
                                    client.Shutdown(SocketShutdown.Both);
                                    client.Close();
                                    client = null;
                                    break;
                                }
                            }

                            // Receive the response from the remote device.
                            while (!_shutdownCamera && !_resetCamera)
                            {
                                try
                                {
                                    // Look for the "magic" number that marks the beginning of an image
                                    if (!Receive(client, 4))
                                    {
                                        //UpdateStatusDescription("rec fail");
                                        break;
                                    }
                                    receiveDone.WaitOne();
                                    if (response == null)
                                    {
                                        //UpdateStatusDescription("rec 1 fail");
                                        break;  // lost connection
                                    }

                                    bool magicNumbersNotFound = true;
                                    bool connectionLost = false;

                                    while (magicNumbersNotFound && !_shutdownCamera && !_resetCamera && !connectionLost)
                                    {
                                        if (ByteArrayCompare(magicNumbers, response))
                                        {
                                            magicNumbersNotFound = false;

                                            // Found start of image

                                            // Read size of image
                                            if (!Receive(client, 4))
                                            {
                                                //UpdateStatusDescription("rec a fail");
                                                connectionLost = true;
                                                break;  // lost connection
                                            }
                                            receiveDone.WaitOne();
                                            if (response == null)
                                            {
                                                //UpdateStatusDescription("rec 2 fail");
                                                connectionLost = true;
                                                break;  // lost connection
                                            }
                                            int imageSize = intFromNetworkBytes(response);

                                            // Read whole image
                                            if (!Receive(client, imageSize))
                                            {
                                                //UpdateStatusDescription("rec b fail");
                                                connectionLost = true;
                                                break;  // lost connection
                                            }
                                            receiveDone.WaitOne();
                                            if (response == null)
                                            {
                                                //UpdateStatusDescription("rec 3 fail");
                                                connectionLost = true;
                                                break;  // lost connection
                                            }

                                            byte[] processedImage; // = new byte[imageSize + huffman_table.Length];
                                            processedImage = AddHuffmanTable(imageSize);
                                            if (processedImage != null)
                                            {
                                                UpdateStatusDescription("At " + DateTime.Now + " got image");
												//File.WriteAllBytes(@"C:\Users\Robotics\Documents\testimage.jpg", processedImage);
																												
												// make sure the image is updated in the UI thread
												Invoke(new UpdateImageDelegate(UpdateImage), processedImage);
                                            }
                                        }
                                        else // response doesn't match magic numbers
                                        {
                                            // Takes the last 3 bytes of the previous response, adds the next byte, and then compares the
                                            // new array to the magic numbers.
                                            prevResponse = response;
                                            byte[] tempResponse = new byte[4];
                                            if (!Receive(client, 1))
                                            {
                                                //UpdateStatusDescription("rec c fail");
                                                connectionLost = true;
                                                break;  // lost connection
                                            }
                                            receiveDone.WaitOne();
                                            if (response == null)
                                            {
                                                //UpdateStatusDescription("rec 4 fail");
                                                connectionLost = true;
                                                break;  // lost connection
                                            }
                                            Array.Copy(prevResponse, 1, tempResponse, 0, 3);
                                            tempResponse[3] = response[0];
                                            response = tempResponse;
                                        }
                                    }
                                }
                                catch (Exception e)
                                {
                                    Console.WriteLine(e.ToString());
                                    UpdateStatusDescription("At " + DateTime.Now + " Failed while finding magic numbers/writing image with error: " + e.ToString());

                                    // break out and start all over looking for the robot
                                    _resetCamera = true;
                                }
                                response = null;
                                //prevResponse = null;

                            } // while

                            if (client != null)
                            {
                                //UpdateStatusDescription("close it");
                                client.Shutdown(SocketShutdown.Both);
                                client.Close();
                                client.Dispose();	// MDF 3/11/2015: Seems unnecessary, but desperate
                                client = null;		// avoid closing twice
                            }
							break;	// get out of foreach loop; no need to waste time
                        } // if
                    } // foreach
                } // try
                catch (Exception e)
                {
                    Console.WriteLine(e.ToString());
                    UpdateStatusDescription("At " + DateTime.Now + " Failed while connecting to socket with error: " + e.ToString());

                    if (client != null)
                    {
                        client.Close();
                        client = null;        //MAB - added to avoid closing twice
                    }

                    // break out and start all over looking for the robot
                    _resetCamera = true;
                }
            }
            //UpdateStatusDescription("At " + DateTime.Now + " DEAD");

            return (ret);
        }

        private byte[] AddHuffmanTable(int imageSize)
        {
            byte[] processedImage = new byte[imageSize + huffman_table.Length];

            // Find if and where we need to insert the huffman table
            bool has_dht = false;
            int pos = 2;
			while (!has_dht && ((pos + 4) <= imageSize))
			{
				if ((response[pos + 1] & 0xff) == 0xc4)
				{
					has_dht = true;
				}
				else if ((response[pos + 1] & 0xff) == 0xda)
				{
					break;
				}

                // Skip to the next marker.
                int marker_size = ((response[pos + 2] & 0xff) << 8) + (response[pos + 3] & 0xff);
                pos += marker_size + 2;
            }

            if (!has_dht && (pos < imageSize)) // MDF 2/24/15: Added test for pos < imageSize, too
            {
                // We need to add huffman table at 'pos' - shift frame data
                // to make room
                try
                {
                    Array.Copy(response, 0, processedImage, 0, pos);
                    Array.Copy(response, pos, processedImage, pos + huffman_table.Length, imageSize - pos);
                    // Now copy in huffman table at 'pos'
                    Array.Copy(huffman_table, 0, processedImage, pos, huffman_table.Length);
                    // Upsize dataLen to include huffman table
                    imageSize += huffman_table.Length;
                }
                catch (Exception e)
                {
                    Console.WriteLine(e.ToString());
                    UpdateStatusDescription("At " + DateTime.Now + " Failed while inserting huffman table with error: " + e.ToString());

                }
            }
            else
            {
                // We will not be inserting the huffman table with this
                // frame
#if !HANDLE_TABLEFUL
                byte[] unProcessedImage = new byte[imageSize];
                Array.Copy (response, unProcessedImage, imageSize);
                return unProcessedImage;
#else
                return null;
#endif
            }

            return (processedImage);
        }

        private bool ByteArrayCompare(byte[] a1, byte[] a2)
        {
            return (StructuralComparisons.StructuralEqualityComparer.Equals(a1, a2));
        }

        private void ConnectCallback(IAsyncResult ar)
        {
            try
            {
                // Retrieve the socket from the state object.
                Socket client = (Socket)ar.AsyncState;

                // Complete the connection.
                client.EndConnect(ar);

                Console.WriteLine("Socket connected to {0}",
                    client.RemoteEndPoint.ToString());

                // Signal that the connection has been made.
                connectError = false;
                connectDone.Set();
            }
            catch (SocketException e)
            {
                Console.WriteLine(e.ToString());
                Console.WriteLine ("Err {0} port {1} ip {2}", e.ErrorCode, _port, _cameraIp);
                UpdateStatusDescription("At " + DateTime.Now + " Failed in ConnectionCallback with error: " + e.ToString());
                connectError = true;
                connectDone.Set();
            }
        }

        private bool Receive(Socket client, int bufferSize)
        {
            bool fSuccess = true;

            try
            {
                // Create the state object.
                StateObject state = new StateObject();
                state.buffer = new Byte[bufferSize];
                state.workSocket = client;

                state.bytesRequested = bufferSize;
                state.bytesReceived = 0;
                receiveDone.Reset();

                // Begin receiving the data from the remote device.
                client.BeginReceive(state.buffer, 0, bufferSize, 0,
                    new AsyncCallback(ReceiveCallback), state);
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToString());
                UpdateStatusDescription("At " + DateTime.Now + " Failed while receiving with error: " + e.ToString());

                fSuccess = false;
            }

            return fSuccess;
        }

        private void ReceiveCallback(IAsyncResult ar)
        {
            try
            {
                // Retrieve the state object and the client socket 
                // from the asynchronous state object.
                StateObject state = (StateObject)ar.AsyncState;
                Socket client = state.workSocket;

                // Read data from the remote device.
                int bytesRead = client.EndReceive(ar);

                if (bytesRead > 0)
                {
                    state.bytesReceived += bytesRead;

                    // There might be more data, so store the data received so far.
                    //state.sb.Append(Encoding.ASCII.GetString(state.buffer, 0, bytesRead));

                    // if not done, yet, get more
                    if (state.bytesReceived < state.bytesRequested)
                    {
                        // Get the rest of the data.
                        client.BeginReceive(state.buffer, state.bytesReceived, (state.bytesRequested - state.bytesReceived), 0,
                            new AsyncCallback(ReceiveCallback), state);
                        return;
                    }
                }

                // All the data has arrived; put it in response.
                //if (state.sb.Length > 1)
                if (state.bytesReceived >= state.bytesRequested)
                {
                    response = state.buffer;
                }
                else
                {
                    response = null;
                }
                // Signal that all bytes have been received.
                receiveDone.Set();
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToString());
                UpdateStatusDescription("At " + DateTime.Now + " Failed in ReceiveCallback with error: " + e.ToString());
                response = null;
                receiveDone.Set();
            }
        }

        private void Send(Socket client, String data)
        {
            // Convert the string data to byte data using ASCII encoding.
            byte[] byteData = Encoding.ASCII.GetBytes(data);

            // Begin sending the data to the remote device.
            client.BeginSend(byteData, 0, byteData.Length, 0,
                new AsyncCallback(SendCallback), client);
        }

        private int intFromNetworkBytes(byte[] byteData)
        {
            int integer = 0;
            integer = integer | (((int)byteData[0] & 0xff) << 24);
            integer = integer | (((int)byteData[1] & 0xff) << 16);
            integer = integer | (((int)byteData[2] & 0xff) << 8);
            integer = integer | (((int)byteData[3] & 0xff));
            return integer;
        }

        private bool Send(Socket client, int[] data)
        {
            try
            {
                // Convert the string data to byte data using ASCII encoding.
                byte[] byteData = new byte[4 * data.Length];
                for (int i = 0; i < data.Length; i++)
                {
                    byteData[(i * 4) + 0] = (byte)((data[i] >> 24) & 255);
                    byteData[(i * 4) + 1] = (byte)((data[i] >> 16) & 255);
                    byteData[(i * 4) + 2] = (byte)((data[i] >> 8) & 255);
                    byteData[(i * 4) + 3] = (byte)(data[i] & 255);
                }

                // Begin sending the data to the remote device.
                client.BeginSend(byteData, 0, byteData.Length, 0,
                    new AsyncCallback(SendCallback), client);
                return (true);
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToString());
                UpdateStatusDescription("At " + DateTime.Now + " Failed while sending with error: " + e.ToString());

                return (false);
            }
        }

        private void SendCallback(IAsyncResult ar)
        {
            try
            {
                // Retrieve the socket from the state object.
                Socket client = (Socket)ar.AsyncState;

                // Complete sending the data to the remote device.
                int bytesSent = client.EndSend(ar);
                Console.WriteLine("Sent {0} bytes to server.", bytesSent);

                // Signal that all bytes have been sent.
                sendError = false;
                sendDone.Set();
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToString());
                UpdateStatusDescription("At " + DateTime.Now + " Failed in SendCallback with error: " + e.ToString());
                sendError = true;
                sendDone.Set();
            }
        }

        private Bitmap byteArrayToImage(byte[] byteArrayIn)
        {
            // make sure we got something
            if (byteArrayIn.Length < 1)
            {
                return (null);
            }
            try
            {
                MemoryStream ms = new MemoryStream(byteArrayIn);
                Bitmap returnImage = new Bitmap(ms);
                //Console.WriteLine("Created Image, size={0}", Image.Size);
                return returnImage;
            }
            catch
            {
                return (null);
            }
        }

        delegate void UpdateStatusDescriptionDelegate(string message);

		private void UpdateStatusDescription(string message)
		{
			if (InvokeRequired)
			{
				BeginInvoke(new UpdateStatusDescriptionDelegate(UpdateStatusDescription), message);
				return;
			}

			if (_statusTextBox != null)
			{
				if (message.StartsWith("At") || (message == ""))
				{
					_statusTextBox.Text = message;
				}
				else
				{
					_statusTextBox.Text = message + "\r\n" + _statusTextBox.Text;
				}
			}
        }

        //MAB - added so the image update happens in the correct thread
        delegate void UpdateImageDelegate(byte[] processedImage);
        private void UpdateImage(byte[] processedImage)
        {
            if ((processedImage == null) || (processedImage.Count() == 1))
            {
                if (!_allowCamera)
                {
                    this.Image = Properties.Resources.CameraDisabled;
                }
                _loadedFile = null; // image is not from the user anymore
                return;
            }

            // id display is paused, then do not show the new image
            if (!_updateDisplay)
            {
                return;
            }

            if (_cameraFlip)
            {
                Bitmap image = byteArrayToImage(processedImage);
                image.RotateFlip(RotateFlipType.Rotate180FlipNone);
                this.Image = image;
            }
            else
            {
                this.Image = byteArrayToImage(processedImage);
            }
            _loadedFile = null; // image is not from the user anymore

            // if we shouldn't do processing, then skip processing.
            // NOTE we do this here so ProcesTarget can be called when there is no robot
            //      and the user loaded their own image; plus it saves calling the method
            if (_visionProcessing || _liveProcessing)
            {
                ProcessTarget();
            }
        }

        // process targets using the currently displayed image
        // file - is supplied is full path of the file name to use for debug images.
        //        set to null so no debug images are created
        private void ProcessTarget(string file = null)
        {
            // valid the image
            if ((this.Image == null) || !(this.Image is Bitmap))
            {
                return;
            }

            int width = this.Image.Width;
            int height = this.Image.Height;
            Bitmap bitmapSrc = this.Image as Bitmap;

            // validate the bitmap
            if ((bitmapSrc == null) || (bitmapSrc.PixelFormat != System.Drawing.Imaging.PixelFormat.Format24bppRgb))
            {
                return;
            }

            // if we didn't create the memory yet OR it is too small, then reallocate it
            // NOTE: we do it this way to save memory allocation time for each image we process
            if ((_rgbValue == null) || (_rgbValue.Length < (width * height)))
            {
                _rgbValue = null;
                _rgbValue = new ThunderVision.RGBValue[width * height];
            }

            // make a copy of the displayed image so we can lock its bits and change them
            // if we need to. we cannot locke the bits of what is being displayed because it messes
            // up what is being shown (i.e. get red X).
            // we need to make a Clone so that the PixelFormat is left untouched
            Rectangle rect = new Rectangle(0, 0, bitmapSrc.Width, bitmapSrc.Height);
            Bitmap bitmap = bitmapSrc.Clone(rect, bitmapSrc.PixelFormat);

            // lock the entire bitmap
            System.Drawing.Imaging.BitmapData bitmapData = bitmap.LockBits(rect, System.Drawing.Imaging.ImageLockMode.ReadWrite, bitmap.PixelFormat);

            // make sure we have enough memory to hold the image
            // NOTE: we do it this way to save memory allocation time for each image we process
            int bytes = Math.Abs(bitmapData.Stride) * bitmap.Height;
            if ((_bitmapRGB == null) || (_bitmapRGB.Length < bytes))
            {
                _bitmapRGB = null;
                _bitmapRGB = new byte[bytes];
            }

            // get the RGB values into an array
            System.Runtime.InteropServices.Marshal.Copy(bitmapData.Scan0, _bitmapRGB, 0, bytes);

            // pull out the RGB values for the target processing and do the processing
            // NOTE: values are in BGR order
            int dst = 0;
            int src;
            for (int h = 0; h < bitmap.Height; h++)
            {
                src = (h * bitmapData.Stride);
                for (int w = 0; w < bitmap.Width; w++)
                {
                    _rgbValue[dst].B = _bitmapRGB[src++];
                    _rgbValue[dst].G = _bitmapRGB[src++];
                    _rgbValue[dst].R = _bitmapRGB[src++];
                    dst++;
                }
            }
            _thunderVision.ProcessImage(_rgbValue, width, height, file);

            // make sure we have someone to call
            // NOTE: if not, we still draw the targets because were called manually by the user)
            if (_visionDelegate == null)
            {
                bitmap.UnlockBits(bitmapData);  // done with bitmap, we are leaving it alone
                bitmap = null;
                return;
            }

            // call delegate
            bool isValid = _visionDelegate(_thunderVision.targetInfo, _thunderVision.targets, width, height);

            // draw the targets on the image
            // NOTE: values are in BGR order
            int rowStride = width * 3;
            int pixel;
            byte r;
            byte g;
            byte b;
            for (int i = 0; i < _thunderVision.targets; i++)
            {
                ThunderVision.TARGET_INFO ti = _thunderVision.targetInfo[i];
                r = 255;
                if (isValid && ti.valid)
                {
                    g = b = 255;
                }
                else
                {
                    g = b = 0;
                }
                for (int x = ti.left; x <= ti.right; x++)           // draw top and bottom
                {
                    pixel = (ti.top * rowStride) + (x * 3);
                    _bitmapRGB[pixel++] = b; _bitmapRGB[pixel++] = g; _bitmapRGB[pixel++] = r;
                    pixel = (ti.bottom * rowStride) + (x * 3);
                    _bitmapRGB[pixel++] = b; _bitmapRGB[pixel++] = g; _bitmapRGB[pixel++] = r;
                }
                for (int y = ti.top; y <= ti.bottom; y++)           // draw left and right
                {
                    pixel = (y * rowStride) + (ti.left * 3);
                    _bitmapRGB[pixel++] = b; _bitmapRGB[pixel++] = g; _bitmapRGB[pixel++] = r;
                    pixel = (y * rowStride) + (ti.right * 3);
                    _bitmapRGB[pixel++] = b; _bitmapRGB[pixel++] = g; _bitmapRGB[pixel++] = r;
                }
            }

            // put back the updated values
            System.Runtime.InteropServices.Marshal.Copy(_bitmapRGB, 0, bitmapData.Scan0, bytes);

            // we are done updating the bitmap
            bitmap.UnlockBits(bitmapData);
            this.Image = bitmap;                // show the updated bitmap
        }

        // handle user changing resolution
        private void _resChoices_SelectedIndexChanged(object sender, EventArgs e)
        {
            int idx = ((_resChoices == null) ? -1 : _resChoices.SelectedIndex);

            switch (idx)
            {
                default:
                    return;     // unsupported value, so return now and do not save the setting
                case 0:
                    SetResolution(CameraControl.Resolution.Resolution640x480);
                    break;
                case 1:
                    SetResolution(CameraControl.Resolution.Resolution320x240);
                    break;
                case 2:
                    SetResolution(CameraControl.Resolution.Resolution160x120);
                    break;
            }

            // save the item
            try
            {
                Properties.Settings.Default.CameraResolution = idx;
                Properties.Settings.Default.Save();
            }
            catch
            {
            }
        }

        // handle user changing frame rate
        private void _fpsChoices_SelectedIndexChanged(object sender, EventArgs e)
        {
            // make sure something is selected
            int idx = ((_fpsChoices == null) ? -1 : _fpsChoices.SelectedIndex);
            if (idx < 0)
            {
                return;
            }
            try
            {
                int fps = Convert.ToInt32((_fpsChoices.SelectedItem as string).Split(' ')[0]);
                SetFrameRate(fps);

                // save the item
                Properties.Settings.Default.CameraFrameRate = fps;
                Properties.Settings.Default.Save();
            }
            catch
            {
            }
        }

        // handle brightness changes
        private void _brightness_ValueChanged(object sender, EventArgs e)
        {
            try
            {
                // make sure the robot knows
                UpdateRobot();

                // save the item
                Properties.Settings.Default.CameraBrightness = _brightness.Value;
                Properties.Settings.Default.Save();
            }
            catch
            {
            }
        }

        // handle exposure changes
        private void _exposure_ValueChanged(object sender, EventArgs e)
        {
            try
            {
                // make sure the robot knows
                UpdateRobot();

                // save the item
                Properties.Settings.Default.CameraExposure = _exposure.Value;
                Properties.Settings.Default.Save();
            }
            catch
            {
            }
        }

        // handle user changing white balance
        private void _whiteBalance_SelectedIndexChanged(object sender, EventArgs e)
        {
            // make sure something is selected
            int idx = ((_whiteBalance == null) ? -1 : _whiteBalance.SelectedIndex);
            if (idx < 0)
            {
                return;
            }
            try
            {
                // make sure the robot knows
                UpdateRobot();

                // save the item
                Properties.Settings.Default.CameraWhiteBalance = (_whiteBalance.SelectedItem as string);
                Properties.Settings.Default.Save();
            }
            catch
            {
            }
        }

        // camera default list
        private void _defaults_Click(object sender, EventArgs e)
        {
            if (DialogResult.Yes != MessageBox.Show("This will set the 2017 defaults.\\n\nAre you sure?", "Set Camera Defaults", MessageBoxButtons.YesNo, MessageBoxIcon.Exclamation, MessageBoxDefaultButton.Button2))
            {
                return;
            }
            _brightness.Value = 20;
            _exposure.Value = -1;
            _whiteBalance.SelectedItem = "ff1";
            _resChoices.SelectedItem = "320 x 240";
            _fpsChoices.SelectedItem = FPS(20);
        }
    
        private string _cameraIp;       // IP Address of camera
        private bool _cameraFlip;       // flip camera image?
		private int _port = 1180;       // port that has image
		private TextBox _statusTextBox; // where to put any status messages (for debugging
        private Resolution _resolution = Resolution.Resolution640x480; // camera resolution
        private int _fps = 30;          // camera frame rate per second
        private volatile bool _shutdownCamera = false;
		private BackgroundWorker _bw;
        private volatile bool _resetCamera = false;
        private bool _started = false;

        // user controls
        private ComboBox _resChoices;                       // list of resolutions
        private ComboBox _fpsChoices;                       // list of frame rates
        private TableLayoutPanel _panelImage;               // panel to hold toolbar buttons for image
        private NumericUpDown _brightness;                  // camera brightness
        private NumericUpDown _exposure;                    // camera exposure
        private ComboBox _whiteBalance;                     // camera white balance
        private Button _defaults;                           // defaults
        private TableLayoutPanel _panelBright;              // panel to hold brightness and exposure
        private ToolTip _tooltip;                           // tooltips for all user controls
        private Button _loadButton;                         // load a bitmap into control
        private Button _enableButton = null;                // stop/start camera connection
        private Button _playButton = null;                  // pause/play live stream
        private Button _saveButton = null;                  // save image that is in control
        private Button _processButton = null;               // process image that is in control
        private Button _liveProcessButton = null;           // live process of streaming video
        private bool _allowCamera = true;                   // stop/start camera connection
        private bool _updateDisplay = true;                 // update image when it comes from robot?
        private string _loadedFile = null;                  // file loaded by user
        private bool _liveProcessing = false;               // do live processing, even if robot doesn't say so

        // for vision processing
        private ThunderVision.RGBValue[] _rgbValue = null;  // buffer for data going to ThunderVision
        private byte[] _bitmapRGB = null;                   // bitmap's RGB values for reading/writing
        private ThunderVision _thunderVision = new ThunderVision();
        private VisionDelegate _visionDelegate = null;
        private bool _visionProcessing = false;

        // ManualResetEvent instances signal completion.
        private ManualResetEvent connectDone = new ManualResetEvent(false);
		private bool connectError = false;
        private bool sendError = false;
        private ManualResetEvent sendDone = new ManualResetEvent(false);
		private ManualResetEvent receiveDone = new ManualResetEvent(false);

		// The response from the remote device.
		private byte[] response;
		private byte[] prevResponse;

		private byte[] magicNumbers = new byte[] { 0x01, 0x00, 0x00, 0x00 };
		private byte[] huffman_table;

		private Socket client;


		private int[] huffman_table_int = new int[] {
			0xFF, 0xC4, 0x01, 0xA2, 0x00, 0x00, 0x01, 0x05, 0x01, 0x01,
			0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00,
			0x00, 0x00, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06,
			0x07, 0x08, 0x09, 0x0A, 0x0B, 0x01, 0x00, 0x03, 0x01,
			0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00,
			0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x02, 0x03, 0x04,
			0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x10, 0x00,
			0x02, 0x01, 0x03, 0x03, 0x02, 0x04, 0x03, 0x05, 0x05,
			0x04, 0x04, 0x00, 0x00, 0x01, 0x7D, 0x01, 0x02, 0x03,
			0x00, 0x04, 0x11, 0x05, 0x12, 0x21, 0x31, 0x41, 0x06,
			0x13, 0x51, 0x61, 0x07, 0x22, 0x71, 0x14, 0x32, 0x81,
			0x91, 0xA1, 0x08, 0x23, 0x42, 0xB1, 0xC1, 0x15, 0x52,
			0xD1, 0xF0, 0x24, 0x33, 0x62, 0x72, 0x82, 0x09, 0x0A,
			0x16, 0x17, 0x18, 0x19, 0x1A, 0x25, 0x26, 0x27, 0x28,
			0x29, 0x2A, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A,
			0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x53,
			0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x63, 0x64,
			0x65, 0x66, 0x67, 0x68, 0x69, 0x6A, 0x73, 0x74, 0x75,
			0x76, 0x77, 0x78, 0x79, 0x7A, 0x83, 0x84, 0x85, 0x86,
			0x87, 0x88, 0x89, 0x8A, 0x92, 0x93, 0x94, 0x95, 0x96,
			0x97, 0x98, 0x99, 0x9A, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6,
			0xA7, 0xA8, 0xA9, 0xAA, 0xB2, 0xB3, 0xB4, 0xB5, 0xB6,
			0xB7, 0xB8, 0xB9, 0xBA, 0xC2, 0xC3, 0xC4, 0xC5, 0xC6,
			0xC7, 0xC8, 0xC9, 0xCA, 0xD2, 0xD3, 0xD4, 0xD5, 0xD6,
			0xD7, 0xD8, 0xD9, 0xDA, 0xE1, 0xE2, 0xE3, 0xE4, 0xE5,
			0xE6, 0xE7, 0xE8, 0xE9, 0xEA, 0xF1, 0xF2, 0xF3, 0xF4,
			0xF5, 0xF6, 0xF7, 0xF8, 0xF9, 0xFA, 0x11, 0x00, 0x02,
			0x01, 0x02, 0x04, 0x04, 0x03, 0x04, 0x07, 0x05, 0x04,
			0x04, 0x00, 0x01, 0x02, 0x77, 0x00, 0x01, 0x02, 0x03,
			0x11, 0x04, 0x05, 0x21, 0x31, 0x06, 0x12, 0x41, 0x51,
			0x07, 0x61, 0x71, 0x13, 0x22, 0x32, 0x81, 0x08, 0x14,
			0x42, 0x91, 0xA1, 0xB1, 0xC1, 0x09, 0x23, 0x33, 0x52,
			0xF0, 0x15, 0x62, 0x72, 0xD1, 0x0A, 0x16, 0x24, 0x34,
			0xE1, 0x25, 0xF1, 0x17, 0x18, 0x19, 0x1A, 0x26, 0x27,
			0x28, 0x29, 0x2A, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A,
			0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x53,
			0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x63, 0x64,
			0x65, 0x66, 0x67, 0x68, 0x69, 0x6A, 0x73, 0x74, 0x75,
			0x76, 0x77, 0x78, 0x79, 0x7A, 0x82, 0x83, 0x84, 0x85,
			0x86, 0x87, 0x88, 0x89, 0x8A, 0x92, 0x93, 0x94, 0x95,
			0x96, 0x97, 0x98, 0x99, 0x9A, 0xA2, 0xA3, 0xA4, 0xA5,
			0xA6, 0xA7, 0xA8, 0xA9, 0xAA, 0xB2, 0xB3, 0xB4, 0xB5,
			0xB6, 0xB7, 0xB8, 0xB9, 0xBA, 0xC2, 0xC3, 0xC4, 0xC5,
			0xC6, 0xC7, 0xC8, 0xC9, 0xCA, 0xD2, 0xD3, 0xD4, 0xD5,
			0xD6, 0xD7, 0xD8, 0xD9, 0xDA, 0xE2, 0xE3, 0xE4, 0xE5,
			0xE6, 0xE7, 0xE8, 0xE9, 0xEA, 0xF2, 0xF3, 0xF4, 0xF5,
			0xF6, 0xF7, 0xF8, 0xF9, 0xFA
		};
	}
}
