﻿// parts of the main form that should not be touched each year
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Net.NetworkInformation;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace ThunderDashboard
{
    public partial class FormMain
    {
        private bool _debug = false;

        private void InitSupport()
        {
            listViewRoboDebug.MouseDoubleClick += ListViewRoboDebug_MouseDoubleClick;
            //controllerIOMap.TabPages.Remove(tabPageCapture);
            groupBoxNetTbl.Visible = _debug;   // hide network table debug groupbox on Settings tab

            comboBoxYear.Items.Add("2017");
            comboBoxYear.Items.Add("2016");
            Properties.Settings.Default.Reload();
            comboBoxYear.SelectedItem = Properties.Settings.Default.Year;
            textBoxTeam.Text = Properties.Settings.Default.Team;

            // set to our default view
            ChangeYear(Properties.Settings.Default.Year, Properties.Settings.Default.Team, true);

            ChartInit();
        }

        // called when the window is first shown
        private void Form_Shown(object sender, EventArgs e)
        {
            // put window in upper-left corner of screen
            Location = new Point(0, 0);
            WindowState = FormWindowState.Normal;   // Driver's Station starts us minimized, so override that

            // sync colors
            SetBackColors();

            // connect to network table
            DashboardNetworkTable.Start(_robotIpAddress, NetworkTableDelegate);

            // if we have a camera, start it
            if (_cameraControl != null)
            {
                _cameraControl.Testing = _debug;
                _cameraControl.Start(_robotIpAddress, _cameraPort, _statusTextBox, _cameraFlip, _cameraVisionDelegate);
            }
        }

        // called when the window is being closed
        private void Form_FormClosed(object sender, FormClosedEventArgs e)
        {
            DashboardNetworkTable.Stop();
            if (_cameraControl != null) { _cameraControl.Stop(); }
        }

        // reset button was pressed
        private void buttonReset_Click(object sender, EventArgs e)
        {
            if (_panelConnected != null) { _panelConnected.BackColor = Color.Yellow; }
            DashboardNetworkTable.Restart();
            if (_cameraControl != null) { _cameraControl.Restart(); }
        }
  
        // handle the debug list being double-clickedre
        private void ListViewRoboDebug_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            DashboardNetworkTable.EditItem(ref listViewRoboDebug, e);
        }

        // called when 'hidden' control to show net tbl debug is double-clicked
        private void panelNetTbl_DoubleClick(object sender, EventArgs e)
        {
            groupBoxNetTbl.Visible = true;
        }

        // see if there are any new controller pics in our directory
        private void CheckControllerPictures(Image driverImage, Image auxImage)
        {
            // they must be in the same directory as our exe
            var location = new Uri(System.Reflection.Assembly.GetEntryAssembly().GetName().CodeBase);
            string directory = new FileInfo(location.AbsolutePath).Directory.FullName;
            string file;

            // check Aux pic
            file = Path.Combine(directory, "Controls Bitmap Aux.PNG");
            if (File.Exists(file))
            {
                try
                {
                    using (FileStream stream = new FileStream(file, FileMode.Open, FileAccess.Read))
                    {
                        auxImage = Image.FromStream(stream);
                        stream.Close();
                    }
                }
                catch { }
            }
            auxControllerPicture.Image = auxImage;

            // check Base pic
            file = Path.Combine(directory, "Controls Bitmap Driver.PNG");
            if (File.Exists(file))
            {
                try
                {
                    using (FileStream stream = new FileStream(file, FileMode.Open, FileAccess.Read))
                    {
                        driverImage = Image.FromStream(stream);
                        stream.Close();
                    }
                }
                catch { }
            }
            driverControllerPicture.Image = driverImage;
        }

        string[] _point = {
            "0.1,10,15,0.3,-0.9",
            "0.2,100,15,0.3,-0.9",
            "0.3,50,15,0.3,-0.9",
            "0.4,300,15,0.3,-0.9",
            "0.6,200,15,0.3,-0.9",
            "0.7,0,15,0.3,-0.9",
            "0.8,50,15,0.3,-0.9",
            "1.5,500,15,0.3,-0.9",
            "1.6,50,15,0.3,-0.9",
            "2,500,15,0.3,-0.9",
            "3,50,15,0.3,-0.9", };
        private void comboBoxYear_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (comboBoxYear.SelectedIndex < 0)
            {
                return;
            }
            switch (comboBoxYear.SelectedItem.ToString())
            {
                default:
                    try
                    {
                        textBoxTeam.Text = comboBoxYear.SelectedItem.ToString().Substring(2) + "15";
                    }
                    catch
                    {

                    }
                    break;

                case "2017":
                    textBoxTeam.Text = "1511";
                    break;
            }
        }

        private void buttonConfiguration_Click(object sender, EventArgs e)
        {
            if (comboBoxYear.SelectedIndex < 0)
            {
                MessageBox.Show("Select a Robot Year");
                return;
            }
            if (textBoxTeam.TextLength != 4)
            {
                MessageBox.Show("Pick a 4 digit team year");
            }
            try
            {
                Properties.Settings.Default.Year = comboBoxYear.SelectedItem.ToString();
                Properties.Settings.Default.Team = textBoxTeam.Text;
                Properties.Settings.Default.Save();
            }
            catch
            {
                return;
            }

            // launch a new one that will use the new settings
            Process process = new Process();
            string file = System.Reflection.Assembly.GetEntryAssembly().GetName().CodeBase;
            process.StartInfo = new ProcessStartInfo(file);
            process.Start();

            // now we can close ourselves
            Close();
        }

        // put PC's IP Address in the network tables
        private void SetPcIpAddress()
        {
            string ip;

            // first try to find the one that matches our team number
            ip = FindIpAddress("10." + _teamNumber.Substring(0, 2) + "." + _teamNumber.Substring(2, 2));
            if (ip == null)
            {
                // didn't find our team network, so take the first one found
                ip = FindIpAddress(null);
            }
            if (ip != null)
            {
                // found an IP to use, so put it in the network table
                DashboardNetworkTable.PutString("thunderdashboard_ip", ip);
            }
        }

        // get our PC's network address for the robot's
        private string FindIpAddress(string startingIp)
        {
            // look at each network card on the PC
            foreach (NetworkInterface n in NetworkInterface.GetAllNetworkInterfaces())
            {
                // make sure the network is connected
                if (n.OperationalStatus != OperationalStatus.Up)
                {
                    continue;       // netowrk is not connected
                }

                // look at each address for the NIC
                foreach (UnicastIPAddressInformation ip in n.GetIPProperties().UnicastAddresses)
                {
                    // make sure it is IPv4
                    if (ip.Address.AddressFamily != System.Net.Sockets.AddressFamily.InterNetwork)
                    {
                        continue;   // not IPv4
                    }

                    // is this the one connected to our robot?
                    if (string.IsNullOrEmpty(startingIp) || ip.Address.ToString().StartsWith(startingIp))
                    {
                        return (ip.Address.ToString());   // this is our robot's network, so this is our PC's address
                    }
                }
            }

            // did not find our robot's network
            return (null);
        }
    }
}
