﻿// For a new year:
// 1. create a new tab, in the designer, and see the "(Name)" property to be tabMainPage followed
//    by the current year; e.g. tabMainPage2018
// 2. copy all the controls from another robot and paste them into the new tab
// 3. you will need to update ALL of the names for the controls. the easiest, and you need to do
//    this for each control, is to copy the name it had on the previous robot's page and update
//    the year to the current year when you paste in that name to the new control
// 4. the Reset event for the buttonReset control must be set to: buttonReset_Click
// 5. 
// 6. in the ChangeYear() method, add a case statement for the current year and fill in as appropriate
// 7. the SetBackColors() method allows you to change the entire background based on robot state; so
//    a update as desired. you may need to add a member variable to help keep track of state
// 8. add new images for the driver and aux controller information
//      - put them in the Images folder, prefixed with the current year
//      - right-click the Images folder in the solution explorer and pick 'Add' -> 'Existing Item...'
//          then browse to the Images folder on the PC to Add them
//      - add them to the Resources.resx (double-click that file in the solution explorer, then
//             draw the images into it.
// 9. in NetworkTableDelegate(), add any new dashboard values for the year.
//    NOTE: be sure to re-use common names:
//      Auto_Mode                       the app sets this to the user selected auto mode (this is a number)
//      thunderdashboard_inpitmode      robot is in pit mode: 1 = yes (used in: 2016)
//      thunderdashboard_gyro           robot's gyro is being calirbrated: 1 = yes (used in: 2016, 2017)
//      thunderdashboard_drivereverse   robot is driving in reversed: 1 = yes (used in: 2016, 2017)
//      thunderdashboard_armlocked      arm is locked into position (used in: 2016, 2017)
//      thunderdashboard_frontcamera    camera feed is the front camera: 1 = yes
//
//      thunderdashboard_breacher       2016 - breacher position form 0 to 100
//      thunderdashboard_scalerpivot    2016 - arm's position from 0 to 100
//      thunderdashboard_scalerext      2016 - arm's extension from 0 to 100
//      thunderdashboard_haveball       2016 - robot has a ball: 1 = yes
//
//      vision_Start                    robot tells us to start/stop vision procesing: true/false (used in: 2017)
//      vision_Counter                  gets incremented each time an image is processed, starting at 1 (used in: 2017)
//      vision_HorizontalOffset         how far off center the entire target is from center of image, in pixels (used in: 2017)
//      vision_DistanceBetween          how far apart are two targets, in pixels (used in: 2017)
//      vision_Result                   0 - no targets at all; 1 - valid target; 2 - not a valid target (used in: 2017)


using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace ThunderDashboard
{
    public partial class FormMain : Form
    {
        // creates the application screens
        public FormMain()
        {
            InitializeComponent();
            InitLights();
            InitSupport();
        }

        // set the screens for the given robot year
        private void ChangeYear(string year, string teamNumber, bool skipStart = false)
        {
            // start with default values
            Text = "Thunder Dashboard " + year + ": " + teamNumber;
            _teamNumber = teamNumber;

            // start by removing all tabs, stop camera, forget about current robot controls
            tabControlDashboard.TabPages.Clear();
            if (!skipStart)
            {
                DashboardNetworkTable.Stop();
                if (_cameraControl != null)
                {
                    _cameraControl.Stop();
                    _cameraControl = null;
                }
            }
            _cameraControl = null;
            _cameraVisionDelegate = null;
            _comboBoxAutoMode = null;
            _autoModeDescriptionsBox = null;
            _panelConnected = null;
            _statusTextBox = null;
            _labelCameraSide = null;
            _labelReverseDrive = null;
            _labelAutoModeValue = null;

            // setup screen based on the year
            switch (year)
            {
                default:
                    MessageBox.Show("The year is not valid.");
                    ChangeYear("2017", "1511");   // pick a valid year so the screen doesn't go blank
                    return;

                case "2017":
                    _cameraPort = 1180;
                    _cameraFlip = true;

                    // not being used yet, so hide them
                    label2017Power.Visible = true;
                    meterControl2017Power.Visible = true;
                    numericUpDown2017Power.Enabled = false;
                    meterControl2017Power.DecimalPlaces(2);

                    // add the tabs we need and set their text
                    tabControlDashboard.TabPages.Add(tabMainPage2017);
                    tabControlDashboard.TabPages.Add(tabPageLightsPsoc42xx);
                    tabControlDashboard.TabPages.Add(tabPageSettings);
                    tabControlDashboard.TabPages.Add(tabPageRoboDebug);
                    tabControlDashboard.TabPages.Add(tabPageChart);
                    tabMainPage2017.Text = "Thunder Shogun"; //TBD: put in robot's official name

                    // set controls to this robot's controls
                    _cameraControl = cameraControl2017;
                    _cameraVisionDelegate = Vision2017;
                    _comboBoxAutoMode = comboBoxAutoMode2017;
                    _autoModeDescriptionsBox = autoModeDescriptionsBox2017;
                    _panelConnected = panelConnected2017;
                    _statusTextBox = statusTextBox2017;
                    _labelCameraSide = labelCameraSide2017;
                    _labelReverseDrive = labelReverseDrive2017;
                    _labelAutoModeValue = labelAutoModeValue2017;

                    // show this years graphics
                    CheckControllerPictures(Properties.Resources._2017_Controls_Bitmap_Driver,
                                            Properties.Resources._2017_Controls_Bitmap_Aux);
                    break;

                case "2016":
                    _cameraPort = 1180;
                    _cameraFlip = false;

                    // add the tabs we need and set their text
                    tabControlDashboard.TabPages.Add(tabMainPage2016);
                    tabControlDashboard.TabPages.Add(tabPageLightsPsoc42xx);
                    tabControlDashboard.TabPages.Add(tabPageSettings);
                    tabControlDashboard.TabPages.Add(tabPageRoboDebug);
                    tabMainPage2016.Text = "Sir Rolling of Thunder";
                    tabPageRoboDebug.Text = "Zack and Kate's Savior";

                    // set controls to this robot's controls
                    _cameraControl = cameraControl;
                    _cameraVisionDelegate = null;
                    _comboBoxAutoMode = comboBoxAutoMode;
                    _autoModeDescriptionsBox = autoModeDescriptionsBox;
                    _panelConnected = panelConnected;
                    _statusTextBox = statusTextBox;
                    _labelCameraSide = labelCameraSide;
                    _labelReverseDrive = labelReverseDrive;
                    _labelAutoModeValue = labelAutoModeValue;

                    // show this years graphics
                    CheckControllerPictures(Properties.Resources._2016_Controls_Bitmap_Driver,
                                            Properties.Resources._2016_Controls_Bitmap_Aux);
                    break;
            }
            _robotIpAddress = "roboRIO-" + _teamNumber + "-FRC.local";

            // hook up with the new autonomous controls
            _autoMode.Initialize(_comboBoxAutoMode, _autoModeDescriptionsBox, _autoModeKey, _keyPrefixAuto);    // tell object which controls to use

            if (!skipStart)
            {
                Form_Shown(null, null);
                // if there is a camera, start it up
                if (_cameraControl != null)
                {
                    _cameraControl.Start(_robotIpAddress, _cameraPort, _statusTextBox, _cameraFlip, _cameraVisionDelegate);
                }
            }
        }

        // set the background colors of controls to indicate what is going on
        private void SetBackColors()
        {
            Color color = Color.Transparent;

            // gyro calibration wins over EVERYTHING
            if (_gyroCalibrate)
            {
                color = Color.Red;
            }
            else if (_inPitMode)
            {
                // when in pit mode we want the color be something 'strange'
                color = Color.PowderBlue;

                if (_colorYellow && !(_colorGreen || _haveBall))
                {
                    color = Color.Plum;                 // warning only
                }
                else if (!_colorYellow && (_colorGreen || _haveBall))
                {
                    color = Color.Orange;
                }
                else if (_colorYellow && (_colorGreen || _haveBall))
                {
                    color = Color.DarkViolet;          // warning and 'good'
                }
            }
            else
            {
                // normal mode
                if (_colorYellow && !(_colorGreen || _haveBall))
                {
                    color = Color.Yellow;               // warning only
                }
                else if (!_colorYellow && (_colorGreen || _haveBall))
                {
                    color = Color.Green;                // 'good' only
                }
                else if (_colorYellow && (_colorGreen || _haveBall))
                {
                    color = Color.YellowGreen;          // warning and 'good'
                }
            }

            // set all the main tab pages, as well as the aux/driver tabs
            foreach (TabPage tabPage in tabControlDashboard.TabPages)
            {
                tabPage.BackColor = color;
            }
            tabPageAuxIOMap.BackColor = color;
            tabPageDriverIOMap.BackColor = color;
        }

        // called when a entry in the table changes
        private void NetworkTableDelegate(string key, string value)
        {
            // validate
            if ((key == null) || (value == null))
            {
                return;
            }

            // make sure the call happens in the UI thread
            if (InvokeRequired)
            {
                BeginInvoke(new DashboardNetworkTable.DashboardNetworkTableDelegate(NetworkTableDelegate), key, value);
                return;
            }

            // see which item it is in the network table that we have
            switch (key)
            {
                default:
                    // see if something we need to send to auto
                    if (key.StartsWith(_keyPrefixAuto))
                    {
                        _autoMode.UpdateValue(key, value);  // update auto variable
                        break;
                    }

                    // see if it is a chart key
                    if (ChartValue(key, value))
                    {
                        break;
                    }

                    // nothing we care about
                    break;

                case DashboardNetworkTable.Connected:
                    // if here, then the robot is there and we should have
                    // connected to the network tables after the robot got there.
                    // so now we need to send down the things that come from the
                    // dashboard so the robot's network table has that info.
                    // if he had done it before the robot's network table was
                    // there, then our setting would only be local and then it
                    // would get wiped out by the robot table.

                    // set default auto mode. MUST do after network table is setup.
                    _autoMode.Connected();
                    if (_panelConnected != null) { _panelConnected.BackColor = Color.Green; }
                    if (_cameraControl != null) { _cameraControl.Connected(); }

                    // tell the robot who we are
                    SetPcIpAddress();
                    break;

                case DashboardNetworkTable.Disconnected:
                    if (_panelConnected != null) { _panelConnected.BackColor = Color.Red; }
                    if (_cameraControl != null) { _cameraControl.Disconnected(); }
                    break;

                case _autoModeKey:
                    if (_labelAutoModeValue != null)
                    {
                        _labelAutoModeValue.Text = value;    // tells us what the robot thinks it is
                    }
                    break;

                case "thunderdashboard_breacher":
                    meterBreacherPosition.Value = value;
                    break;

                case "thunderdashboard_scalerpivot":
                    meterScalerPivot.Value = value;
                    break;

                case "thunderdashboard_scalerext":
                    meterScalerExtension.Value = value;
                    break;

                case "thunderdashboard_inpitmode":
                    _inPitMode = (value == "1");
                    SetBackColors();    // make sure background colors stay in sync
                    break;

                case "thunderdashboard_gyro":
                    _gyroCalibrate = (value == "1");
                    SetBackColors();    // make sure background colors stay in sync
                    break;

                case "thunderdashboard_yellow":
                    _colorYellow = (value == "1");
                    SetBackColors();    // make sure background colors stay in sync
                    break;

                case "thunderdashboard_green":
                    _colorGreen = (value == "1");
                    SetBackColors();    // make sure background colors stay in sync
                    break;

                case "thunderdashboard_haveball":
                    _haveBall = (value == "1");
                    SetBackColors();    // make sure background colors stay in sync
                    break;

                case "thunderdashboard_drivereverse":
                    // sees if robot drive is reversed
                    if (_labelReverseDrive != null)
                    {
                        if (value == "1")
                        {
                            _labelReverseDrive.Visible = true;
                        }
                        else
                        {
                            _labelReverseDrive.Visible = false;
                        }
                    }
                    break;

                case "thunderdashboard_armlocked":
                    // sees if scaler arm is locked and changes color based on that
                    if (value == "1")
                    {
                        meterScalerPivot.Color(Color.Red);
                    }
                    else
                    {
                        meterScalerPivot.ColorReset();
                    }
                    break;

                case "thunderdashboard_frontcamera":
                    if (_labelCameraSide != null)
                    {
                        if (value == "1")
                        {
                            _labelCameraSide.Text = "F\r\nR\r\nO\r\nN\r\nT";
                        }
                        else
                        {
                            _labelCameraSide.Text = "R\r\nE\r\nA\r\nR";
                        }
                    }
                    break;

                case "shooter_desiredSpeed":
                    meterControl2017Power.Value = value;
                    break;
                case "shooter_desiredSpeed_e":
                    try
                    {
                        double speed = Convert.ToDouble(value);
                        if (speed != Convert.ToDouble(numericUpDown2017Power.Value))
                        {
                            _speed = speed;
                            numericUpDown2017Power.Value = Convert.ToDecimal(speed);
                        }
                    }
                    catch
                    {
                    }
                    break;

                case "vision_Start":
                    if (_cameraControl == null)
                    {
                        break;      // no camera control
                    }
                    if (value != "true")
                    {
                        DashboardNetworkTable.PutNumber("vision_Result", (double)0);    // make robot doesn't read anything
                        _cameraControl.VisionStop();
                    }
                    else
                    {
                        _cameraControl.VisionStart();
                    }
                    break;
            }

            // update the robot debug table
            if (!DashboardNetworkTable.ShowTableItem(ref listViewRoboDebug, key, value))
            {
                if (_statusTextBox != null)
                {
                    _statusTextBox.Text = DashboardNetworkTable.LastError();
                }
            }
        }

        //
        //      vision_Result                   0 - no targets at all; 1 - valid target; 2 - not a valid target (used in: 2017)
        private bool Vision2017(ThunderVision.TARGET_INFO[] targetInfo, int targets, int imageWidth, int imageHeight)
        {
            int distanceBetween = 0;
            int leftCenter = 0;
            int rightCenter = 0;
            int horizontalOffset = 0;
            int result = 0;
            List<ThunderVision.TARGET_INFO> validTargets = new List<ThunderVision.TARGET_INFO>();

            // if camera is looking too far down, we might get a flare from the gear
            // (reflection from the yellow can look green). but when that is the case,
            // the two field targets will be higher in the image.
            // if we have a gear flare, we should only ever have one flare; this would
            // be because the camera is not centered within the gap between the two gear
            // teeth and the flare would be off the tooth closest to the camera.
            // however, we need to be careful for the case of the camera looking too
            // far up and the two field targets are touching the bottom of the image.
            //
            // start by seeing how many are touching the bottom
            int onBottom = 0;
            for (int i = 0; i < targets; i++)
            {
                if (targetInfo[i].bottom >= (imageHeight - 10))
                {
                    onBottom++;
                }
            }
            // now process the targets
            for (int i = 0; i < targets; i++)
            {
                // if only one on the bottom and this is it, then skip it
                if ((onBottom == 1) && (targetInfo[i].bottom >= (imageHeight - 10)))
                {
                    targetInfo[i].valid = false;        // make it as being invalid
                }
                else if (targetInfo[i].top < 10)
                {
                    targetInfo[i].valid = false;        // make it as being invalid
                }
                else
                {
                    validTargets.Add(targetInfo[i]);    // keep it because it is within the image
                }
            }

            // see if the targets are valid
            targets = validTargets.Count;
            if (targets < 1)
            {
                result = 0;                     // no targets found
            }
            else if (targets != 2)
            {
                result = 2;                     // we need two targets
            }
            else if (validTargets[0].row != validTargets[1].row)
            {
                result = 2;                     // they must be next to each other
            }
            else
            {
                // we have valid targets
                result = 1;

                // calc based on which one is to the real left
                if (validTargets[0].left <= validTargets[1].left)
                {
                    // first target is on the left
                    distanceBetween = validTargets[1].left - validTargets[0].right;
                    leftCenter = validTargets[0].left + (validTargets[0].maxWidth / 2);
                    rightCenter = validTargets[1].left + (validTargets[1].maxWidth / 2);
                }
                else
                {
                    // first target is on the right
                    distanceBetween = validTargets[0].left - validTargets[1].right;
                    leftCenter = validTargets[1].left + (validTargets[1].maxWidth / 2);
                    rightCenter = validTargets[0].left + (validTargets[0].maxWidth / 2);
                }
                horizontalOffset = ((rightCenter - leftCenter) / 2) + leftCenter;
                horizontalOffset = (horizontalOffset - (imageWidth / 2));
            }

            DashboardNetworkTable.PutNumber("vision_DistanceBetween", (double)distanceBetween);
            DashboardNetworkTable.PutNumber("vision_HorizontalOffset", (double)horizontalOffset);
            DashboardNetworkTable.PutNumber("vision_Result", (double)result);
            DashboardNetworkTable.PutNumber("vision_Counter", (double)_visionCounter++);
            return (result == 1);
        }


        private string _robotIpAddress = null;
        private string _teamNumber = "1511";
        private int _cameraPort = 1180;
        private bool _cameraFlip = false;

        private const string _autoModeKey = "Auto_Mode";
        private const string _keyPrefixAuto = "thunderdashboard_auto_";
        private AutoMode _autoMode = new AutoMode();

        // vision
        private int _visionCounter = 1;                 // increment for each processed image

        // the current robot tab controls
        private CameraControl _cameraControl = null;    // camera fed
        private CameraControl.VisionDelegate _cameraVisionDelegate = null;  // callback when vision has processed the image
        private Panel _panelConnected = null;           // connection status
        private ComboBox _comboBoxAutoMode = null;      // autonomous choices
        private TextBox _autoModeDescriptionsBox = null;// autonomous description
        private TextBox _statusTextBox = null;          // status text
        private Label _labelCameraSide = null;          // front/back camera status
        private Label _labelReverseDrive = null;        // forward/reverse status
        private Label _labelAutoModeValue = null;       // autonomous number

        private bool _inPitMode = false;
        private bool _gyroCalibrate = false;
        private bool _colorYellow = false;
        private bool _colorGreen = false;

        // 2016 specific
        private bool _haveBall = false;

        private void numericUpDown2017Power_ValueChanged(object sender, EventArgs e)
        {
            try
            {
                double speed = Convert.ToDouble(numericUpDown2017Power.Value);
                if (speed != _speed)
                {
                    _speed = speed;
                    DashboardNetworkTable.PutNumber("shooter_desiredSpeed_e", _speed);
                }
            }
            catch
            {
            }
        }
        private double _speed = 0;

        private void checkBox2017Power_CheckedChanged(object sender, EventArgs e)
        {
            numericUpDown2017Power.Enabled = checkBox2017Power.Checked;
        }
    }
}